from kallikrein import Expectation

from amino import Map

from ribosome.plugin import Config, PluginSettings
from ribosome.machine import message
from ribosome.machine.state import SubTransitions, AutoData
from ribosome.machine import trans
from ribosome.machine.messages import Stage1
from ribosome.settings import RequestHandlers, RequestHandler, Plain
from ribosome.nvim import NvimIO
from ribosome.test.integration.klk import AutoPluginIntegrationKlkSpec

Msg = message('Msg')


class Env(AutoData):
    pass


class Core(SubTransitions):

    @trans.unit(Stage1)
    def stage_1(self) -> None:
        pass

    @trans.unit(Msg, trans.nio)
    def msg1(self) -> NvimIO[None]:
        return NvimIO(lambda v: v.vars.set('success', 1))


class APSettings(PluginSettings):
    pass


auto_config = Config(
    name='plug',
    prefix='plug',
    state_type=Env,
    plugins=Map(core=Core),
    settings=APSettings(),
    request_handlers=RequestHandlers.cons(
        RequestHandler.msg_cmd(Msg)('msg', prefix=Plain, sync=True)
    ),
)


class AutoPluginSpec(AutoPluginIntegrationKlkSpec[APSettings, Env]):
    '''
    zero-setup plugin $auto_plugin
    '''

    @property
    def _prefix(self) -> str:
        return 'plug'

    def module(self) -> str:
        return __name__

    def config_name(self) -> str:
        return 'auto_config'

    def _pre_start(self) -> None:
        super()._pre_start()
        self.vim.vars.set_p('components', ['core'])

    def auto_plugin(self) -> Expectation:
        self.vim.cmd_once_defined('PlugStage1')
        self.cmd_sync('Msg')
        return self.var_becomes('success', 1)

__all__ = ('AutoPluginSpec',)



from amino import Map

from typing import Callable, Any

from ribosome.record import Record, field, maybe_field
from ribosome.nvim import NvimFacade, AsyncVimProxy


class Data(Record):
    vim_facade = maybe_field((NvimFacade, AsyncVimProxy))
    sub_states = field(Map, initial=Map())

    def sub_state_m(self, name):
        return self.sub_states.get(name)

    def sub_state(self, name, default: Callable[[], Any]):
        return self.sub_state_m(name) | default

    def with_sub_state(self, name, state):
        new_states = self.sub_states + (name, state)
        return self.set(sub_states=new_states)

__all__ = ('Data',)



import sys
import abc
import typing
from typing import Union, Any, Callable, Type, Optional, TypeVar, Generic, GenericMeta

import neovim

from amino import List

import ribosome.nvim.components
from ribosome.nvim import NvimFacade
from ribosome.machine import Message
from ribosome.logging import nvim_logging, Logging
from ribosome.request import msg_command, msg_function, command, function, json_msg_command
from ribosome.machine.base import ShowLogInfo
from ribosome.machine.scratch import Mapping
from ribosome.rpc import rpc_handlers_json
from ribosome.record import encode_json
from ribosome.machine.messages import UpdateState, Stage1
from ribosome.machine.state import AutoRootMachine, AutoData, RootMachineBase
from ribosome.settings import Config, PluginSettings, Full, Short


class NvimPluginBase(Logging):
    name: Optional[str] = None
    debug: bool = False

    @property
    def plugin_name(self) -> str:
        return type(self).name or 'ribosome'

    def __init__(self, nvim: Union[NvimFacade, neovim.Nvim]) -> None:
        self.vim = NvimFacade(nvim, self.plugin_name) if isinstance(nvim, neovim.Nvim) else nvim
        self.setup_logging()

    def setup_logging(self) -> None:
        self.file_log_handler = nvim_logging(self.vim)


NB = TypeVar('NB', bound=NvimPluginBase)
N = TypeVar('N', bound='NvimPlugin')
NM = TypeVar('NM', bound='NvimPluginMeta')


class NvimPluginMeta(GenericMeta):

    def __new__(
            cls: Type[NM],
            name: str,
            bases: tuple,
            namespace: dict,
            pname: str=None,
            prefix: str=None,
            debug: bool=False
    ) -> Type[N]:
        inst: Type[N] = super().__new__(cls, name, bases, namespace)
        if pname is not None:
            setup_plugin(inst, pname, prefix or pname, debug)
        return inst


class NvimPlugin(NvimPluginBase, metaclass=NvimPluginMeta):
    prefix: Optional[str] = None

    def stage_1(self) -> None:
        pass

    def stage_2(self) -> None:
        pass

    def stage_3(self) -> None:
        pass

    def stage_4(self) -> None:
        pass

    def quit(self) -> None:
        pass

    def rpc_handlers(self) -> List[str]:
        return rpc_handlers_json(type(self))

    def set_log_level(self, level: str) -> None:
        self.file_log_handler.setLevel(level)

    def append_python_path(self, path: str) -> None:
        sys.path.append(path)

    def show_python_path(self) -> typing.List[str]:
        return sys.path

    @neovim.autocmd('VimLeave', sync=True)
    def vim_leave(self) -> None:
        ribosome.nvim.components.shutdown = True
        self.quit()


NSP = TypeVar('NSP', bound='NvimStatePlugin')
NSPM = TypeVar('NSPM', bound='NvimStatePluginMeta')


class NvimStatePluginMeta(NvimPluginMeta):

    def __new__(
            cls: Type[NSPM],
            name: str,
            bases: tuple,
            namespace: dict,
            pname: str=None,
            prefix: str=None,
            debug: bool=False
    ) -> Type[NSP]:
        inst: Type[NSP] = super().__new__(cls, name, bases, namespace, pname, prefix, debug)
        if inst.name is not None:
            setup_state_plugin(inst, inst.name, inst.prefix or inst.name, debug)
        return inst


class NvimStatePlugin(NvimPlugin, metaclass=NvimStatePluginMeta):

    @abc.abstractmethod
    def state(self) -> RootMachineBase:
        ...

    def message_log(self) -> List[Message]:
        return self.state().message_log // encode_json

    def state_data(self) -> str:
        return self.state().data.json | '{}'

    def plug_command(self, plug_name: str, cmd_name: str, *args: str) -> None:
        self.state().plug_command(plug_name, cmd_name, args)


Settings = TypeVar('Settings', bound=PluginSettings)
D = TypeVar('D', bound=AutoData)
APM = TypeVar('APM', bound='AutoPluginMeta')
AP = TypeVar('AP', bound='AutoPlugin')


class AutoPluginMeta(NvimStatePluginMeta):

    def __new__(
            cls: Type[APM],
            name: str,
            bases: tuple,
            namespace: dict,
            pname: str=None,
            prefix: str=None,
            debug: bool=False,
            config: Config[Settings, D]=None,
    ) -> Type['AutoPlugin[Settings, D]']:
        inst: Type[AutoPlugin[Settings, D]] = super().__new__(cls, name, bases, namespace, pname, prefix, debug)
        if config is not None:
            setup_auto_plugin(inst, config)
        return inst


class AutoPlugin(Generic[Settings, D], NvimStatePlugin, metaclass=AutoPluginMeta):

    def __init__(self, nvim: Union[NvimFacade, neovim.Nvim], config: Config[Settings, D]) -> None:
        super().__init__(nvim)
        self.config = config
        self.root = self.create_root()

    def create_root(self) -> AutoRootMachine[Settings, D]:
        title = self.plugin_name
        return AutoRootMachine(self.vim, self.config, title)

    def stage_1(self) -> None:
        self.root.start()
        self.root.wait_for_running()
        self.root.send(Stage1())

    def state(self) -> RootMachineBase:
        return self.root


class Helpers(Logging):

    def __init__(self, cls: type, name: str, prefix: str) -> None:
        self.cls = cls
        self.name = name
        self.prefix = prefix

    def handler(self, name: str, dec: Callable[..., Callable], handler: Callable, *a: Any, **kw: Any) -> None:
        def wrap(self: Any, *a: Any, **kw: Any) -> Any:
            try:
                return handler(self, *a, **kw)
            except Exception as e:
                self.log.caught_exception_error(f'calling handler `{name}`', e)
        func = dec(*a, name=name, **kw)(wrap)
        setattr(self.cls, name, func)

    def name_handler(self, suf: str, dec: Callable[..., Callable], handler: Callable, *a: Any, **kw: Any) -> None:
        n = f'{self.name}_{suf}'
        self.handler(n, dec, handler, *a, **kw)

    def short_handler(self, suf: str, dec: Callable[..., Callable], handler: Callable, *a: Any, **kw: Any) -> None:
        n = f'{self.prefix}_{suf}'
        self.handler(n, dec, handler, *a, **kw)

    def msg_cmd(self, suf: str, msg: type) -> None:
        self.short_handler(suf, msg_command, lambda: None, msg)

    def msg_fun(self, suf: str, msg: type) -> None:
        self.short_handler(suf, msg_function, lambda: None, msg)

    def json_msg_cmd(self, suf: str, msg: type) -> None:
        self.short_handler(suf, json_msg_command, lambda: None, msg)


def setup_plugin(cls: Type[NvimPlugin], name: str, prefix: str, debug: bool) -> None:
    help = Helpers(cls, name, prefix)
    cls.name = name
    cls.prefix = prefix
    cls.debug = debug
    help.msg_cmd('show_log_info', ShowLogInfo)
    help.short_handler('log_level', command, cls.set_log_level)
    help.msg_fun('mapping', Mapping)
    help.name_handler('stage_1', command, cls.stage_1, sync=True)
    help.name_handler('stage_2', command, cls.stage_2, sync=True)
    help.name_handler('stage_3', command, cls.stage_3, sync=True)
    help.name_handler('stage_4', command, cls.stage_4, sync=True)
    help.name_handler('quit', command, cls.quit, sync=True)
    help.name_handler('rpc_handlers', function, cls.rpc_handlers, sync=True)
    help.name_handler('append_python_path', function, cls.append_python_path)
    help.name_handler('show_python_path', function, cls.show_python_path)


def setup_state_plugin(cls: Type[NSP], name: str, prefix: str, debug: bool) -> None:
    help = Helpers(cls, name, prefix)
    help.short_handler('state', function, cls.state_data)
    help.json_msg_cmd('update_state', UpdateState)
    help.short_handler('plug', command, cls.plug_command)
    if debug:
        help.name_handler('message_log', function, cls.message_log, sync=True)


def setup_auto_plugin(cls: Type[AutoPlugin], config: Config[Settings, D]) -> None:
    help = Helpers(cls, config.name, config.prefix)
    for hname, handler in config.request_handlers.handlers.items():
        dispatcher = handler.dispatcher
        helper = (
            help.short_handler
            if handler.prefix is Short else
            help.name_handler
            if handler.prefix is Full else
            help.handler
        )
        helper(handler.name, dispatcher.decorator(), lambda: None, *dispatcher.args)

__all__ = ('NvimPlugin', 'NvimStatePlugin')



import abc
import inspect
from typing import Callable, Type, TypeVar, Generic, Iterable, cast, Union, Any
from enum import Enum

from amino import List, Either, Lists, Map, _, L, __, Nil
from amino.func import flip
from amino.util.string import ToStr

from ribosome.nvim import NvimIO
from ribosome.nvim.components import NvimComponent
from ribosome.logging import Logging
from ribosome.data import Data
from ribosome.machine.message_base import Message
from ribosome.machine.transitions import Transitions
from ribosome.request import command, function, msg_command, json_msg_command, msg_function


A = TypeVar('A', contravariant=True)
B = TypeVar('B')


class PluginSetting(Generic[A, B], ToStr):

    def __init__(self, name: str, desc: str, help: str, prefix: bool, tpe: Type[A], ctor: Callable[[A], B]) -> None:
        self.name = name
        self.desc = desc
        self.help = help
        self.prefix = prefix
        self.tpe = tpe
        self.ctor = ctor

    @property
    def value(self) -> NvimIO[Either[str, A]]:
        def read(v: NvimComponent) -> Either[str, A]:
            vars = v.vars
            getter = vars.p if self.prefix else vars
            return vars.typed(self.tpe, getter(self.name))
        return NvimIO(read)

    def value_or(self, default: A) -> NvimIO[A]:
        return self.value / __.value_or(default)

    def _arg_desc(self) -> List[str]:
        return List(self.name, str(self.prefix), str(self.tpe))


def setting_ctor(tpe: Type[A], ctor: Callable[[A], B]) -> Callable[[str, str, str, bool], PluginSetting[A, B]]:
    def setting(name: str, desc: str, help: str, prefix: bool) -> PluginSetting[A, B]:
        return PluginSetting(name, desc, help, prefix, tpe, ctor)
    return setting


class PluginSettings(ToStr, Logging):

    def __init__(self) -> None:
        self.components = list_setting('components', 'names or paths of active components', components_help, True)

    def all(self) -> Map[str, PluginSetting]:
        settings = inspect.getmembers(self, L(isinstance)(_, PluginSetting))
        return Map(Lists.wrap(settings).map(_[1]).apzip(_.name).map2(flip))

    def _arg_desc(self) -> List[str]:
        return List(str(self.all()))


components_help = '''The plugin can run an arbitrary set of sub-components that inherit the class `Component`.
They receive all messages the core of the plugin processes and have access to the plugin state.
They can define nvim request handlers with the decorator `trans.xyz`.
The entries of this setting can either be names of the builtin components or arbitrary python module paths that define
custom plugins.
'''
str_setting = setting_ctor(str, lambda a: a)
list_setting = setting_ctor(list, cast(Callable[[Iterable[A]], List[B]], Lists.wrap))

M = TypeVar('M', bound=Message)


class RequestDispatcher(ToStr):

    @abc.abstractmethod
    def decorator(self) -> Callable[..., Callable]:
        ...

    @abc.abstractproperty
    def args(self) -> List[Any]:
        ...


class MsgDispatcher(Generic[M], RequestDispatcher):

    def __init__(self, msg: Type[M]) -> None:
        self.msg = msg

    def _arg_desc(self) -> List[str]:
        return List(str(self.msg))

    @property
    def args(self) -> List[Any]:
        return List(self.msg)


class MsgCmd(Generic[M], MsgDispatcher[M]):

    def decorator(self) -> Callable[..., Callable]:
        return msg_command


class JsonMsgCmd(Generic[M], MsgDispatcher[M]):

    def decorator(self) -> Callable[..., Callable]:
        return json_msg_command


class MsgFun(Generic[M], MsgDispatcher[M]):

    def decorator(self) -> Callable[..., Callable]:
        return msg_function


class PrefixStyle(ToStr):

    def _arg_desc(self) -> List[str]:
        return Nil


class Short(PrefixStyle):
    pass


class Full(PrefixStyle):
    pass


class Plain(PrefixStyle):
    pass


class RequestHandler(ToStr):

    def __init__(self, dispatcher: RequestDispatcher, name: str, prefix: PrefixStyle=Short, sync: bool=False) -> None:
        self.dispatcher = dispatcher
        self.name = name
        self.prefix = prefix
        self.sync = sync

    @staticmethod
    def msg_cmd(msg: Type[M]) -> 'RequestHandlerBuilder':
        return RequestHandlerBuilder(MsgCmd(msg))

    @staticmethod
    def msg_fun(msg: Type[M]) -> 'RequestHandlerBuilder':
        return RequestHandlerBuilder(MsgFun(msg))

    def _arg_desc(self) -> List[str]:
        return List(str(self.dispatcher), self.name, str(self.prefix), str(self.sync))


class RequestHandlerBuilder:

    def __init__(self, dispatcher: RequestDispatcher) -> None:
        self.dispatcher = dispatcher

    def __call__(self, name: str, prefix: bool=True, sync: bool=False) -> RequestHandler:
        return RequestHandler(self.dispatcher, name, prefix, sync)


class RequestHandlers(ToStr):

    @staticmethod
    def cons(*handlers: RequestHandler) -> 'RequestHandlers':
        return RequestHandlers(Map(Lists.wrap(handlers).apzip(_.name).map2(flip)))

    def __init__(self, handlers: Map[str, RequestHandler]) -> None:
        self.handlers = handlers

    def _arg_desc(self) -> List[str]:
        return List(str(self.handlers))


Settings = TypeVar('Settings', bound=PluginSettings)
S = TypeVar('S', bound=Data)
T = TypeVar('T', bound=Transitions)


class Config(Generic[Settings, S], ToStr):

    def __init__(self, name: str, prefix: str, plugins: Map[str, Union[str, Type[T]]], state_type: Type[S],
                 settings: Settings, request_handlers: RequestHandlers) -> None:
        self.name = name
        self.prefix = prefix
        self.plugins = plugins
        self.state_type = state_type
        self.settings = settings
        self.request_handlers = request_handlers

    def _arg_desc(self) -> List[str]:
        return List(str(self.plugins), str(self.settings), str(self.request_handlers))

__all__ = ('PluginSetting', 'setting_ctor', 'PluginSettings', 'str_setting', 'list_setting', 'Config',
           'RequestHandlers', 'RequestHandler')
